open Utils
open Logic_ast
open Comp_ast
open Logic_core
open Logic_typing

let case_kind f1 f2 f3 ~case:(_,t1) = 
  match t1 with
      CHol -> f1
    | CCtx -> f2
    | _ -> f3

let flatten_all_cterm t =
  cterm_map ~cholterm:(fun t -> CHolTerm(flatten_all_lmodal t))
            ~cctxterm:(fun t -> CCtxTerm(flatten_all_lctxdesc t)) t

let normalize_ctx e = 
  match e with
      LCtxAsList(ctx) ->
  	let norm_ctx = List.map (fun (a,b) -> (a,lterm_map ~lmodal:(fun _ _ mt1 subst -> modal_apply_subst mt1 subst) b)) ctx in
  	  LCtxAsList(norm_ctx)
    | _ -> e

let normalize_modal t =
  match t with
       LTermInCtx(ctx,tm) ->
  	 let norm_tm = lterm_map ~lmodal:(fun _ _ mt1 subst -> modal_apply_subst mt1 subst) tm in
  	   nowarn let LCtxAsList(norm_ctx) = normalize_ctx (LCtxAsList(ctx)) in
  	     LTermInCtx(norm_ctx, norm_tm)
    | _ -> t

let normalize_cterm t =
  cterm_map ~cholterm:(fun t -> CHolTerm(normalize_modal t))
            ~cctxterm:(fun t -> CCtxTerm(normalize_ctx t)) t

let remove_infer_cterm =
  cterm_map ~cholterm:(fun t -> CHolTerm(remove_infer_lmodal t))
            ~cctxterm:(fun t -> CCtxTerm(remove_infer_lctxdesc t))
            ~cinfer:(fun l f rf ->
	      (match !(match_cunif l) with
                  Inst(t,_) -> rf t
                | Uninst(_,t) ->
		  let ts = match !t with
		            Some(l) -> (match !l with
				         Inst(t,_) -> ("failure to infer hole of type: " ^ Comp_print.string_of_cterm (normalize_cterm (rf t)))
			               | _ -> "")
		          | None -> ""
		  in
		  print_string (ts ^ "\n");
		  flush stdout;
		  failwith ("couldn't infer comp. unification variable!")))

let subst_bound_ctx subst term =
  let term' = CtxbindCtermS.subst_bound subst term in
  flatten_all_cterm term'

let rec elim_inst_cunif e =
  match e with
      CInfer(l,f) -> (match !(match_cunif l) with Inst(t,_) -> elim_inst_cunif (f t) | _ -> e)
    | _ -> e

let subst_bound ~case:t1 ~term:t2 ~subst:t3 =
  case_kind
    (function CHolTerm(t) -> MetabindCtermS.subst_bound t | _ -> failwith "problem in subst_bound")
    (function CCtxTerm(t) -> subst_bound_ctx t | _ -> failwith "problem in subst_bound")
    CbindCtermS.subst_bound
    ~case:t1
    (elim_inst_cunif t3)
    t2

let get_cdef_term n cdefenv = let (_, tm, _, _) = ExtDict.find n cdefenv in tm
let get_cdef_type n cdefenv = let (tp, _, _, _) = ExtDict.find n cdefenv in tp

let flatten_closure (metasubst,ctxsubst,compsubst) e =
  flatten_all_cterm
    (MetabindCtermS.subst_bound_list (List.rev metasubst)
       (CtxbindCtermS.subst_bound_list (List.rev ctxsubst)
	  (CbindCtermS.subst_bound_list (List.rev compsubst) e)))
    
let rec ctype_whnf e =
  match elim_inst_cunif e with
      CApp(t1, t3) ->
	(match ctype_whnf t1 with
	     CLambda(s, t1, t2) ->
	       ctype_whnf (subst_bound ~case:s ~term:t2 ~subst:t3)
	   | t1' -> CApp(t1', t3))
    | CClosure(subst, e) ->
	ctype_whnf (flatten_closure subst e)
    | t -> t

let rec ctype_whdelta cdefenv e =
  monadic option_monad in begin
    let rec aux e =
      match e with
	  CNVar(n) -> cmd { return (get_cdef_term n cdefenv) }
	| CApp(e1, e2) ->  cmd { e1' <- aux e1 then
				 return (CApp(e1', e2)) }
	| e -> None
    in
      cmd { e' <- aux e then
	    return (ctype_whnf e') }
  end

let rec ctype_fullwhdelta cdefenv e =
  let e' = ctype_whnf e in
    match ctype_whdelta cdefenv e with
	None -> e'
      | Some e'' -> ctype_fullwhdelta cdefenv e''


let add_to_metaenv tp (defenv, cdefenv, metaenv, ctxenv, compenv) =
  let metaenv' = tp :: metaenv in
    (defenv, cdefenv, metaenv', ctxenv, compenv)

let add_to_ctxenv tp (defenv, cdefenv, metaenv, ctxenv, compenv) =
  let ctxenv' = () :: ctxenv in
    (defenv, cdefenv, metaenv, ctxenv', compenv)

let add_to_compenv tp (defenv, cdefenv, metaenv, ctxenv, compenv) =
  let compenv' = tp :: compenv in
    (defenv, cdefenv, metaenv, ctxenv, compenv')

let add_to_env ~case:var t =
  case_kind
    (function CHolTerm(t) -> add_to_metaenv t | _ -> failwith "add_to_env with HOL binder and non-HOL term")
    (function CSort(CCtx) -> add_to_ctxenv () | _ -> failwith "add_to_env with CTX binder and non-CTX term")
    (function CHolTerm(_) | CSort(CCtx) -> failwith "add_to_env with COMP binder and non-COMP term" | t -> add_to_compenv t)
    ~case:var
    (elim_inst_cunif t)
let env_len       ~case:var (_,_,metaenv,ctxenv,compenv) = case_kind (List.length metaenv) (List.length ctxenv) (List.length compenv) ~case:var
let open_up       ~case:var ?(howmany=1) env t =
  let n = env_len ~case:var env in
    case_kind (MetabindCterm.open_up ~howmany:howmany n)
              (CtxbindCterm.open_up  ~howmany:howmany n)
              (CbindCterm.open_up ~howmany:howmany n)
              ~case:var
              t
let close_down    ~case:var ?(howmany=1) env t =
  let n = env_len ~case:var env in
    case_kind (MetabindCterm.close_down ~howmany:howmany n)
              (CtxbindCterm.close_down  ~howmany:howmany n)
              (CbindCterm.close_down ~howmany:howmany n)
              ~case:var
              t
let has_bound_var = case_kind MetabindCterm.has_bound_var CtxbindCterm.has_bound_var CbindCterm.has_bound_var

let open_up_2 ~case1:((_,var1) as s1) ~case2:((_,var2) as s2) env =
  if var1 = var2 then open_up ~case:s1 ~howmany:2 env else compose (open_up ~case:s1 env) (open_up ~case:s2 env)

let close_down_2 ~case1:((_,var1) as s1) ~case2:((_,var2) as s2) env =
  if var1 = var2 then close_down ~case:s1 ~howmany:2 env else compose (close_down ~case:s1 env) (close_down ~case:s2 env)

let add_to_envn =
  case_kind
    (fun (defenv,cdefenv,metan,ctxn,compn) -> (defenv,cdefenv,metan+1,ctxn,compn))
    (fun (defenv,cdefenv,metan,ctxn,compn) -> (defenv,cdefenv,metan,ctxn+1,compn))
    (fun (defenv,cdefenv,metan,ctxn,compn) -> (defenv,cdefenv,metan,ctxn,compn+1))
let open_up_n ~case:var ?(howmany=1) (_,_,metan,ctxn,compn) t =
  let n = case_kind ~case:var metan ctxn compn in
    case_kind (MetabindCterm.open_up ~howmany:howmany n)
              (CtxbindCterm.open_up  ~howmany:howmany n)
              (CbindCterm.open_up ~howmany:howmany n)
              ~case:var
              t

let unified (t : cterm * cterm) = ()

let ctype_equal_n (defenv,cdefenv,metan,ctxn,compn) ctype_whnf e1 e2 =
  let rec auxfull ((_,_,metan,ctxn,compn) as envn) e1 e2 = 
    let aux = auxfull envn in
    let aux_under v t1 t2 = let envn' = add_to_envn ~case:v envn in auxfull envn' (open_up_n ~case:v envn t1) (open_up_n ~case:v envn t2) in
    let e1' = ctype_whnf e1 in
    let e2' = ctype_whnf e2 in
    match e1', e2' with
	CSort(s1), CSort(s2) -> s1 = s2
      | CPi(v, a1, b1), CPi(_, a2, b2) ->
	aux a1 a2 && aux_under v b1 b2
      | CLambda(v, a1, b1), CLambda(_, a2, b2) ->
	aux a1 a2 && aux_under v b1 b2
      | CApp(a1, b1), CApp(a2, b2) when aux a1 a2 && aux b1 b2 -> true
      | CHolTerm(t1), CHolTerm(t2) -> lmodal_equal defenv metan t1 t2
      | CCtxTerm(t1), CCtxTerm(t2) -> lctxdesc_equal defenv t1 t2
      | CUnitType, CUnitType -> true
      | CSigma(v, a1, b1), CSigma(_, a2, b2) ->
	aux a1 a2 && aux_under v b1 b2
      | CBVar(i1), CBVar(i2) -> i1 = i2
      | CFVar(i1), CFVar(i2) -> i1 = i2
      | CProdType(a1,b1), CProdType(a2,b2) -> aux a1 a2 && aux b1 b2
      | CRecType(v,a1,b1), CRecType(_,a2,b2) ->
	aux a1 a2 && aux_under v b1 b2
      | CSumType(b1), CSumType(b2) -> List.for_all2 aux b1 b2
      | CRefType(t1), CRefType(t2) -> aux t1 t2
      | CIntType, CIntType -> true
      | CStringType, CStringType -> true
      | CBoolType, CBoolType -> true
      | CArrayType(t1), CArrayType(t2) -> aux t1 t2
      | CInfer(l1,f1), CInfer(l2,f2) when match_cunif l1 == match_cunif l2 -> true
      | CInfer(l1,f1), CInfer(l2,f2) ->
	(let l1 = match_cunif l1 in
	 let l2 = match_cunif l2 in
	 cunifvar_type_unify aux l1 f1 l2 f2;
	 match !l1, !l2 with
	     Inst(t1,otyp1), _ -> aux (f1 t1) e2'
	   | _, Inst(t2,otyp2) -> aux e1' (f2 t2)
	   | (Uninst(_,otyp1)) , (Uninst(_,otyp2)) ->
	     l1 := Inst(CInfer(l2,fun x -> x),otyp2); true)

      | CInfer(l,f), e
      | e, CInfer(l,f) ->
	  (let l = match_cunif l in
	     match !l with
	       Inst(t,_) -> aux (f t) e
	     | Uninst(i,otyp) ->
		 (l := Inst(e,otyp);
		  (unified (f e, e);
		   (* try aux e1' e2'
		   with _ -> failwith "can't unify" *)
		   aux e1' e2')))
      | CNVar(s1), CNVar(s2) when s1 = s2 -> true
      | CNVar(s1), e2 when aux (get_cdef_term s1 cdefenv) e2 -> true
      | e1, CNVar(s2) when aux e1 (get_cdef_term s2 cdefenv) -> true
      | e1, e2 -> (match ctype_whdelta cdefenv e1, ctype_whdelta cdefenv e2 with
		       Some(e1), Some(e2) -> aux e1 e2
		     | Some(e1), None -> aux e1 e2
		     | None, Some(e2) -> aux e1 e2
		     | None, None -> false)
  in
    auxfull (defenv,cdefenv,metan,ctxn,compn) e1 e2

let ctype_equal_n env e1 e2 = (ctype_equal_n env (fun t -> t) e1 e2) || (ctype_equal_n env ctype_whnf e1 e2)

let ctype_equal (defenv,cdefenv,metaenv,ctxenv,compenv) = ctype_equal_n (defenv,cdefenv,List.length metaenv,List.length ctxenv,List.length compenv)

let ctx_is_prefix defenv l1 l2 =
    if (List.length l1 > List.length l2) then false
    else ExtList.foldindex (fun i ((_,a),(_,b)) res -> res && (lterm_equal defenv i a b)) true
      (List.combine l1 (ExtList.take (List.length l1) l2))

let ctype_sub ((defenv,cdefenv,metaenv,ctxenv,compenv) as env) e1 e2 =

  let modal_subtype (defenv,_,metaenv,_,_) e1 e2 =
    match e1, e2 with
  	LTermInCtx(ct1,t1), LTermInCtx(ct2,t2) ->
  	  if ctx_is_prefix defenv ct1 ct2 then
	    lmodal_equal defenv (List.length metaenv) (LTermInCtx(ct2,BindLterm.shift_bound (List.length ct2 - List.length ct1) t1)) (LTermInCtx(ct2,t2))
  	  else
  	    false
      | _, _ -> false
  in

  let rec subaux ((defenv,cdefenv,metan,ctxn,compn) as envn) e1 e2 =
    let (<:) = subaux envn in
    let sub_under v t1 t2 = 
      let env' = add_to_envn ~case:v envn in
      subaux env' (open_up_n ~case:v envn t1) (open_up_n ~case:v envn t2)
    in
    if ctype_equal_n envn e1 e2 then true
    else (match ctype_whnf e1, ctype_whnf e2 with
	      CPi(v, a1, b1), CPi(_, a2, b2) -> a2 <: a1 && sub_under v b1 b2
	    | CHolTerm(t1), CHolTerm(t2) -> modal_subtype env t1 t2

	    | CSigma(v, a1, b1), CSigma(_, a2, b2) when ctype_equal_n envn a1 a2 && sub_under v b1 b2 -> true
	    | CSigma(v, (CHolTerm(LTermInCtx(ct1,t1)) as a1), b1), CSigma(_, CHolTerm(LTermInCtx(ct2,t2)), b2) ->
		if ctx_is_prefix defenv ct1 ct2 && lmodal_equal defenv metan (LTermInCtx(ct2,BindLterm.shift_bound (List.length ct2 - List.length ct1) t1)) (LTermInCtx(ct2,t2)) then
		  let b2' = MetabindCtermS.subst_bound (LTermInCtx(ct1,LModal(LBMeta 0,boundlist (List.length ct1)))) b2 in
		    sub_under v b1 b2'
		else
		  false
	    | CProdType(a1, b1), CProdType(a2, b2) -> a1 <: a2 && b1 <: b2
	    | CRecType(v, a1, b1), CRecType(_, a2, b2) -> ctype_equal_n envn a1 a2 && sub_under v b1 b2
	    | CSumType(b1), CSumType(b2) -> List.for_all2 (<:) b1 b2
	    | CNVar(s1), e2 when (get_cdef_term s1 cdefenv) <: e2 -> true
	    | e1, CNVar(s2) when e1 <: (get_cdef_term s2 cdefenv) -> true
	    | e1, e2 -> (match ctype_whdelta cdefenv e1, ctype_whdelta cdefenv e2 with
			     Some(e1), Some(e2) -> e1 <: e2
			   | Some(e1), None -> e1 <: e2
			   | None, Some(e2) -> e1 <: e2
			   | None, None -> false))
  in	    
    subaux (defenv, cdefenv, List.length metaenv, List.length ctxenv, List.length compenv) e1 e2


let rec ctype_appmany_open t =
  match t with
      CApp(t1, t2) -> (let base, args = ctype_appmany_open t1 in
			 base, List.append args [t2])
    | _ -> t, []

let rec ctype_appmany t l =
  List.fold_left (fun cur elm -> CApp(cur,elm)) t l
	 

(* SPECIFICATION OF THE LANGUAGE *)
let ctype_spec_abstractions =
  [ (CType, CType);
    (CCtx,  CType);
    (CHol,  CType);
    (CKind, CType);
    (CKind, CKind);
    (CHol,  CKind);
    (CCtx,  CKind)]

let ctype_spec_existentials =
  [ (CHol, CType);
    (CCtx, CType) ]

let rec is_decreasing_bound curmax subst =
  match subst with
      [] -> true
    | LVar (LBVar hd) :: tl -> hd = curmax - 1 && is_decreasing_bound hd tl
    | _ -> false

let subst_fmeta scrutinee subst tm =
  let i = match scrutinee with 
               CHolTerm(LFMeta(i)) -> Some i
             | CHolTerm(LTermInCtx(ctx, LModal(LFMeta(i), subst))) when is_decreasing_bound (List.length ctx) subst -> Some i
	     | _ -> None
  in
  match i with
      Some i ->
	(nowarn let CHolTerm(t) = subst in
	 flatten_all_cterm (MetabindCterm.subst_fvar t i tm))
    | _ -> tm

let subst_fmeta_in_env ((defenv, cdefenv, metaenv, ctxenv, compenv) as env) scrutinee t =
  let i = match scrutinee with 
               CHolTerm(LFMeta(i)) -> Some i
             | CHolTerm(LTermInCtx(ctx, LModal(LFMeta(i), subst))) when is_decreasing_bound (List.length ctx) subst -> Some i
	     | _ -> None
  in
  match i with
      Some i ->
	(nowarn let CHolTerm(t) = t in
	 let metaenv' = List.map (compose flatten_all_lmodal (MetabindModal.subst_fvar t i)) metaenv in
	 let compenv' = List.map (compose flatten_all_cterm (MetabindCterm.subst_fvar t i)) compenv in
	   (defenv, cdefenv, metaenv', ctxenv, compenv'))
    | _ -> 
	env

let subst_fctx scrutinee subst tm =
  match scrutinee with
      CCtxTerm(LFCtx(i))
    | CCtxTerm(LCtxAsList([_,LTermList(LFCtx(i))])) ->
	(nowarn let CCtxTerm(t) = subst in
	 flatten_all_cterm (CtxbindCterm.subst_fvar t i tm))
    | _ -> tm

let subst_fctx_in_env ((defenv, cdefenv, metaenv, ctxenv, compenv) as env) i t =
  match i with
      CCtxTerm(LFCtx(i))
    | CCtxTerm(LCtxAsList([_,LTermList(LFCtx(i))])) ->
	(nowarn let CCtxTerm(t) = t in
	 let metaenv' = List.map (compose flatten_all_lmodal (CtxbindModal.subst_fvar t i)) metaenv in
	 let compenv' = List.map (compose flatten_all_cterm (CtxbindCterm.subst_fvar t i)) compenv in
	   (defenv, cdefenv, metaenv', ctxenv, compenv'))
    | _ -> env

let many_add_to_env vars ts env =
  List.fold_left (fun env (var, t) -> add_to_env ~case:var t env) env (List.combine vars ts)

let rec type_of_cterm ((defenv, cdefenv, metaenv, ctxenv, compenv) as env) (e : cterm) =
  match e with
      CSort(CType) -> CSort(CKind)
    | CSort(CKind) -> failwith "Kind does not have a classifier"
    | CSort(CCtx)  -> failwith "ctx is not part of the cterm language"
    | CSort(CHol)  -> failwith "HOL is not part of the cterm language"
    | CApp(CNVar(magic), t) when magic="magic" ->
	(match type_of_HOL env t with
	     CHolTerm(LTermInCtx(_, LSort(LProp))) -> CSigma((None,CHol), t, CUnitType)
	   | _ -> failwith "magic used wrongly")
    | CPi(var, t1, t2) ->
	(let t1_s = csort_of ~case:var env t1 in
	 let env' = add_to_env ~case:var t1 env in
	 let t2'  = open_up ~case:var env t2 in
	 let t2_s = csort_of_cterm env' t2' in
	   if List.mem (t1_s, t2_s) ctype_spec_abstractions then
	     CSort(t2_s)
	   else
	     failwith "non-allowed abstraction used")
    | CLambda(var, t1, e2) ->
	(let env' = add_to_env ~case:var t1 env in
	 let t2   = type_of_cterm env' (open_up ~case:var env e2) in
	 let tp   = CPi(var, t1, close_down ~case:var env' t2) in
	 let _    = type_of_cterm env tp in
	   tp)
    | CApp(e1, e2) ->
	(let t1 = ctype_fullwhdelta cdefenv (type_of_cterm env e1) in
	   match t1 with
	       CPi(var, t, t') ->
		 let t2 = type_of ~case:var env e2 in
		   if ctype_sub env t2 t then
		     subst_bound ~case:var ~term:t' ~subst:e2
		   else
		     failwith "types don't match in application"
	     | _ -> failwith "applying something to a non-functional type")
    | CHolTerm(t) ->
	failwith "HOL terms are not part of the cterm language"
    | CCtxTerm(t) ->
	failwith "ctx terms are not part of the cterm language"
    | CUnitType -> CSort(CType)
    | CUnitExpr -> CUnitType
    | CSigma(var, t1, t2) ->
	(let t1_s = csort_of ~case:var env t1 in
	 let env' = add_to_env ~case:var t1 env in
	 let t2'  = open_up ~case:var env t2 in
	 let t2_s = csort_of_cterm env' t2' in
	   if List.mem (t1_s, t2_s) ctype_spec_existentials then
	     CSort(t2_s)
	   else
	     failwith "non-allowed existential type used")
    | CPack(t1, var, ret, e) ->
	(let t1_t = type_of ~case:var env t1 in
	 let tp = CSigma(var, t1_t, ret) in
	 let _ = type_of_cterm env tp in
	 let e_t = type_of_cterm env e in
	 let e_expected = subst_bound ~case:var ~term:ret ~subst:t1 in
	   if ctype_sub env e_t e_expected then
	     tp
	   else
	     failwith "body of existential doesn't have expected type")
    | CUnpack(t1, var1, var2, e) ->
	(let t1_t = ctype_fullwhdelta cdefenv (type_of_cterm env t1) in
	   match t1_t with
	       CSigma(_, tt1, tt2) ->
		 let env' = add_to_env ~case:var1 tt1 env in
		 let tt2' = open_up ~case:var1 env tt2 in
		 let env'' = add_to_env ~case:var2 tt2' env' in
		 let e' = open_up_2 ~case1:var1 ~case2:var2 env e in
		 let e_t = close_down_2 ~case1:var1 ~case2:var2 env'' (type_of_cterm env'' e') in
		   if has_bound_var ~case:var1 0 e_t then
		     failwith "witness should not escape from existential unpacking"
		   else if csort_of_cterm env e_t = CType then
		     e_t
		   else
		     failwith "unpacking inside non-computational term"
	     | _ -> failwith "unpacking a non-existential term")
    | CBVar(i) -> failwith "type_of_cterm on a bound variable!"
    | CFVar(i) -> List.nth compenv (List.length compenv - i - 1)
    | CNVar(s) -> get_cdef_type s cdefenv

    | CHolCase(ts,vars,ret,branches) ->
	(let check_one_scrutinee t =
	   let t_t = type_of_HOL env t in
	   let _ =
	     match t_t with
		 CHolTerm(LTermInCtx(_, LSort(_))) -> ()
	       | _ -> (match type_of_HOL env t_t with
			   CHolTerm(LTermInCtx(_, LSort(_))) -> ()
			 | _ ->failwith "cannot pattern match on proof objects!")
	   in
	     t_t
	 in
	
	 let ts_t = List.map check_one_scrutinee ts in
	 let env' = many_add_to_env vars ts_t env in
	 let open_in_env' = open_up ~case:(None, CHol) ~howmany:(List.length ts) env in

	 let ret' = open_in_env' ret in
	 let scrutinees = ts in

	 let env'' =
	   List.fold_left
	     (fun env' (i, scrutinee) ->
		subst_fmeta_in_env env' (open_in_env' scrutinee) (CHolTerm(LFMeta(List.length metaenv + i))))
	     env' (List.combine (increasing (List.length ts)) scrutinees)
	 in

	 let ret_s = csort_of_cterm env'' ret' in
	 let ret_f ts = MetabindCtermS.subst_bound_list (List.map (function CHolTerm(t) -> t | _ -> failwith "ret_f") ts) ret in
	   if ret_s = CType then
	     (let _ = List.iter (check_branch env ts_t scrutinees ret_f) branches in
		ret_f ts)
	   else
	     failwith "matching on HOL terms is only allowed for computational terms")

    | CCtxCase(t,var,ret,branches) ->
	(let t_t = type_of_ctx env t in
	 let env' = add_to_env ~case:var t_t env in
	 let ret' = open_up ~case:var env ret in

	 let scrutineevar = t in
	 let env'' = subst_fctx_in_env env' (open_up ~case:var env scrutineevar) (CCtxTerm(LFCtx(env_len ~case:var env))) in

	 let ret_s = csort_of_cterm env'' ret' in
	 let ret_f t = subst_bound ~case:var ~term:ret ~subst:t in
	   if ret_s = CType then
	     (let _ = List.iter (check_ctxbranch env scrutineevar ret_f) branches in
		ret_f t)
	   else
	     failwith "matching on context terms is only allowed for computational terms")

    | CProdType(t1,t2) ->
	(if ctype_equal env (type_of_cterm env t1) (CSort(CType)) &&
	    ctype_equal env (type_of_cterm env t2) (CSort(CType)) then
	   CSort(CType)
	 else
	   failwith "product type with non-product terms")
    | CTuple(e1,e2) ->
	(let t1 = type_of_cterm env e1 in
	 let t2 = type_of_cterm env e2 in
	 let tp = CProdType(t1,t2) in
	 let _ = type_of_cterm env tp in
	   tp)
    | CProj(i,e) ->
	(let t = ctype_fullwhdelta cdefenv (type_of_cterm env e) in
	   match t with
	       CProdType(t1,t2) ->
		 (if i = 1 then t1
		  else if i = 2 then t2
		  else failwith "proj with invalid index")
	     | _ -> failwith "proj on non-tuple")
    | CRecType(var, k, t) ->
	(let _ = if csort_of_cterm env k = CKind then () else failwith "recursive type's kind is not CKind" in
	 let env' = add_to_env ~case:var k env in
	 let t' = open_up ~case:var env t in
	   if ctype_equal env' (type_of_cterm env' t') (open_up ~case:var env k) then
	     k
	   else
	     failwith "recursive type's definition does not match expected kind")
    | CFold(e1, t2) ->
	(let t1 = type_of_cterm env e1 in
	 let _ = type_of_cterm env t2 in
	 let t2base, t2args = ctype_appmany_open (ctype_fullwhdelta cdefenv t2) in
	   match ctype_fullwhdelta cdefenv t2base with
	     CRecType(var, k, t) ->
	       (let t1exp = ctype_appmany (subst_bound ~case:var ~subst:t2base ~term:t) t2args in
		  if ctype_sub env t1 t1exp then
		    t2
		  else
		    failwith "fold applied to invalid instantiation of recursive type")
	     | _ ->
		 failwith "fold with non-recursive type")
    | CUnfold(e1) ->
	(let t1 = type_of_cterm env e1 in
	 let t1base, t1args = ctype_appmany_open (ctype_fullwhdelta cdefenv t1) in
	   match ctype_fullwhdelta cdefenv t1base with
	       CRecType(var, k, t) ->
		 ctype_appmany (subst_bound ~case:var ~subst:t1base ~term:t) t1args
	     | _ ->
		 failwith "unfold with non-recursive type argument")
    | CSumType(branches) ->
	(let checkbranch branch = ctype_equal env (type_of_cterm env branch) (CSort CType) in
	   if List.for_all checkbranch branches then
	     CSort(CType)
	   else
	     failwith "there is a branch with non-computational type")
    | CCtor(i,t,par) ->
	(let t' = ctype_fullwhdelta cdefenv t in
	 let par_t = type_of_cterm env par in
	   match t' with
	       CSumType(branches) ->
		 let branch_t = List.nth branches i in
		   if ctype_sub env par_t branch_t then 
		     t
		   else
		     failwith "parameter of constructor is not of right type"
	     | _ -> failwith "constructor used with non-sum type")
    | CMatch(e,branches) ->
	(let e_t = type_of_cterm env e in
	 let constrs_t = match ctype_fullwhdelta cdefenv e_t with CSumType(b) -> b | _ -> failwith "match on term of non-sum type" in
	   (* this is a hack. take care of it *)
	 let t = 
	   let constr_t = List.nth constrs_t 1 in
	   let (var,body) = List.nth branches 1 in
	   let env' = add_to_env ~case:var constr_t env in
	   let body' = open_up ~case:var env body in
	   let body_t = close_down ~case:var env' (type_of_cterm env' body') in
	     body_t
	 in
	 let checkbranch (var,body) constr_t =
	   let env' = add_to_env ~case:var constr_t env in
	   let body' = open_up ~case:var env body in
	   let body_t = close_down ~case:var env' (type_of_cterm env' body') in
	     if not (ctype_equal env body_t t) then failwith "branch doesn't match expected type"
	 in
	 let _ = List.iter2 checkbranch branches constrs_t in
	 let _ = if not (ctype_equal env (type_of_cterm env t) (CSort(CType))) then failwith "match with non-computational type as result" in
	   t)
    | CLetRec(defs, e) ->
	(let n = List.length defs in
	 let envlen = List.length compenv in
	 let envlen' = n + envlen in
	 let check_def_type (_,t,_) = if not (ctype_equal env (type_of_cterm env t) (CSort(CType))) then failwith "letrec with non-computational type of definition" in
	 let _ = List.iter check_def_type defs in
	 let env' = List.fold_left (fun env (var,t,_) -> add_to_env ~case:var t env) env defs in
	 let check_def (_,t,e) =
	   let e' = CbindCterm.open_up ~howmany:n envlen e in
	   let e_t = CbindCterm.close_down ~howmany:n envlen' (type_of_cterm env' e') in
	     if not (ctype_equal env e_t t) then failwith "a definition in letrec has wrong type!"
	 in
	 let _ = List.iter check_def defs in
	 let e_t = CbindCterm.close_down ~howmany:n envlen' (type_of_cterm env' (CbindCterm.open_up ~howmany:n envlen e)) in
	   if ctype_equal env (type_of_cterm env e_t) (CSort CType) then
	     e_t
	   else
	     failwith "letrec returning non-computational term")
    | CLet(v,d,e) ->
	(let d_t = type_of ~case:v env d in
	 let e_t = 
	   if ctype_equal env (type_of_cterm env d_t) (CSort(CType)) then
	     (let env' = add_to_env ~case:v d_t env in
		close_down ~case:v env' (type_of_cterm env' (open_up ~case:v env e)))
	   else
	     (type_of_cterm env (subst_bound ~case:v ~subst:d ~term:e))
	 in
	      if ctype_equal env (type_of_cterm env e_t) (CSort CType) then
		e_t
	      else
		failwith "let returning non-computational term")

    | CRefType(t) ->
	(let _ = check_computational env t in CSort(CType))
    | CMkRef(e,t') ->
	(let t = type_of_cterm env e in
	 let _ = check_computational env t in
	   if ctype_equal env t t' then
	     CRefType(t)
	   else
	     failwith "designated ref type doesn't match the value"
	)
    | CAssign(e1,e2) ->
	(let t1 = ctype_fullwhdelta cdefenv (type_of_cterm env e1) in
	 let t2 = type_of_cterm env e2 in
	   match t1 with
	       CRefType(t) ->
		 if ctype_equal env t2 t then
		   CUnitType
		 else
		   failwith "non type-compatible assignment"
	     | _ -> 
		 failwith "assignment to non-reference")
    | CReadRef(e) ->
	(let t = ctype_fullwhdelta cdefenv (type_of_cterm env e) in
	   match t with
	       CRefType(t) -> t
	     | _ -> failwith "dereferencing a non-reference")
    | CLoc(l,t) -> CRefType(t)
    | CSeq(e1,e2) ->
	(ensure_type_is env CUnitType e1;
	 let t2 = type_of_cterm env e2 in
	 let _ = check_computational env t2 in
	   t2)

    | CIntType -> CSort(CType)
    | CIntConst(i) -> CIntType
    | CStringType -> CSort(CType)
    | CStringConst(s) -> CStringType
    | CIntOp(o,e1,e2) ->
	(ensure_type_is env CIntType e1;
	 ensure_type_is env CIntType e2;
	 CIntType)
    | CIntTest(o,e1,e2) ->
	(ensure_type_is env CIntType e1;
	 ensure_type_is env CIntType e2;
	 CBoolType)
    | CBoolType -> CSort(CType)
    | CBoolConst(b) -> CBoolType
    | CBoolOp(o,e1,e2) ->
	(ensure_type_is env CBoolType e1;
	 ensure_type_is env CBoolType e2;
	 CBoolType)
    | CIfThenElse(e1,e2,e3) ->
	(ensure_type_is env CBoolType e1;
	 let t2 = type_of_cterm env e2 in
	 let t3 = type_of_cterm env e3 in
	   check_computational env t2;
	   if ctype_equal env t2 t3 then
	     t2
	   else
	     failwith "types of if-branches do not match")

    | CArrayType(t) ->
	check_computational env t;
	CSort(CType)

    | CArrayLit(ts,t) ->
	(let _ = List.iter (fun e -> if not (ctype_equal env (type_of_cterm env e) t) then failwith "array literal's members do not agree on type") ts in
	   CArrayType(t))

    | CMkArray(e1,e2,t_exp) ->
	  ensure_type_is env CIntType e1;
	  check_computational env t_exp;
	  ensure_type_is env t_exp e2;
	  CArrayType(t_exp)

    | CArrayGet(e1,e2) ->
	(let t1 = ctype_fullwhdelta cdefenv (type_of_cterm env e1) in
	  ensure_type_is env CIntType e2;
	  match t1 with
	      CArrayType(t) -> t
	    | _ -> failwith "array type expected when indexing array")

    | CArraySet(e1,e2,e3) ->
	(let t1 = ctype_fullwhdelta cdefenv (type_of_cterm env e1) in
	  ensure_type_is env CIntType e2;
	  match t1 with
	      CArrayType(t) ->
		if ctype_equal env (type_of_cterm env e3) t then
		  CUnitType
		else
		  failwith "assigning different-type value to array index"
	    | _ -> failwith "array type expected when indexing array")

    | CArrayLen(e) ->
	(let t = ctype_fullwhdelta cdefenv (type_of_cterm env e) in
	  match t with
	      CArrayType(_) -> CIntType
	    | _ -> failwith "array type expected when using arraylen")

    | CArrayLoc(_,t) -> CArrayType(t)

    | CHolHash(e) ->
	(let t = type_of_cterm env e in
	   check_computational env t;
	   CIntType)

    | CPrint(e) ->
	(let t = type_of_cterm env e in
	   check_computational env t;
	   CUnitType)

    | CClosure(subst, e) ->
	(let (_,_,metaenv',ctxenv',compenv') as env' = env_from_subst defenv cdefenv subst in
	 let (<.>) = compose in
	 let e' = ((MetabindCterm.open_up ~howmany:(List.length metaenv') 0) <.>
	            (CtxbindCterm.open_up ~howmany:(List.length ctxenv') 0) <.>
		    (CbindCterm.open_up ~howmany:(List.length compenv') 0)) e in
	  type_of_cterm env' e')

    | CTypeAscribe(e, t) ->
	(let t' = type_of_cterm env e in
	   if ctype_equal env t t' then
	     t'
	   else
	     failwith "type ascription does not match type")

    | CPreEval(tac,ctx,params) ->
	(let cappmany base lst = List.fold_left (fun cur elm -> CApp(cur,elm)) base lst in
	 let check_ctx ((defenv,_,_,_,_) as env) comparison ctx params =
	   (nowarn let CCtxTerm(LCtxAsList(ctx)) = ctx in
	    let paramsT = List.map (type_of_HOL env) params in
	    let _ = List.iter
	      (fun elm ->
		 nowarn let CHolTerm(LTermInCtx(ctx',_)) = elm in
		   if not (comparison defenv ctx' ctx) then
		     failwith "comparison of ctx to current meta's context failed")
	      paramsT
	    in
	      ())
	 in
	 let check_ctx_prefix env ctx params = check_ctx env ctx_is_prefix ctx params in
	 let check_ctx_same   env ctx params = check_ctx env (fun d l1 l2 -> lctxdesc_equal d (LCtxAsList(l1)) (LCtxAsList(l2))) ctx params in

	 let rec is_decreasing_bound curmax subst =
	   match subst with
	       [] -> true
	     | LVar (LBVar hd) :: tl -> hd = curmax - 1 && is_decreasing_bound hd tl
	     | _ -> false
	 in
	 let ctxN = (nowarn let CCtxTerm(LCtxAsList(ctx)) = ctx in List.length ctx) in
	 let check_id_substs t =
	   nowarn let CHolTerm(LTermInCtx(_,t')) = t in
	   let _ = lterm_map ~lmodal:(fun _ subst rt1 rtsubst ->
	 			if is_decreasing_bound ctxN subst then
	 			  LModal(rt1,rtsubst)
	 			else
	 			  failwith "metavars used with non-identity substs")
	     t'
	   in
	     ()
	 in

	 let _ = type_of_ctx env ctx in
	 let _ = List.iter (fun t -> ignore(type_of_HOL env t)) params in
	 let e' = cappmany tac (ctx :: params) in
	 let t_t = type_of_cterm env e' in
	 let prop = (match ctype_fullwhdelta cdefenv t_t with
			 CSigma(_, prop, CUnitType) ->
			   (match type_of_HOL env prop with
				CHolTerm(LTermInCtx(_, LSort(LProp))) -> prop
			      | _ -> failwith "tactic does not return proof object")
		       | _ -> failwith "preeval application to wrong tactic type") in
	 let _ = (match tac with CNVar(_) -> () | _ -> failwith "preeval of expression instead of named function") in
	 let _ = check_ctx_same   env ctx params  in
	 let _ = check_ctx_same   env ctx [prop]  in
	 let metaenvcterm = (List.map (fun x -> CHolTerm(x)) metaenv) in
	 let _ = check_ctx_prefix env ctx metaenvcterm in
	 (* let _ = List.iter check_id_substs (List.append (prop :: params) metaenvcterm) in *)
	 t_t)

    | CInfer(el, f) ->
	(match !(match_cunif el) with
	     Inst(ei,ti) ->
	       (let t' = type_of_cterm env (f ei) in
		  if (match !ti with Some ot -> ctype_equal env t' (CInfer(ot,f)) | _ -> true) then
		    t'
		  else
		    failwith "instantiation of inferred term with wrong type!")
	   | Uninst(i,otyp) ->
	       (match !otyp with
		    None -> (let uv = mk_cunifvar () in otyp := (Some uv); CInfer(uv, f))
		  | Some ti -> CInfer(ti, f)))
	               
	
and ensure_type_is ((defenv,cdefenv,_,_,_) as env) t e = 
  if not (ctype_equal env (type_of_cterm env e) t) then failwith "expected another type"
    
and check_computational ((defenv,cdefenv,_,_,_) as env) t =

  if not (ctype_equal env (type_of_cterm env t) (CSort(CType))) then failwith "computational type expected"

and check_pattern (defenv, _, _, _, _) metas t =

  (let ctx, tm = match t with CHolTerm(LTermInCtx(ctx,tm)) -> ctx, tm | _ -> failwith "patterns should be terms-in-context" in
     Logic_typing.check_pattern defenv ctx (List.length metas) tm)
	 
and check_branch ((defenv, cdefenv, metaenv, ctxenv, compenv) as env) typs scrutinees ret_f (pats, e) =
  
  let open_up_meta n = MetabindCterm.open_up ~howmany:n (List.length metaenv) in

  let replace_scrutinees_in_env open_up env scrutinees ts =
    List.fold_left
      (fun env (scrutinee, t) -> subst_fmeta_in_env env (open_up scrutinee) t)
      env (List.combine scrutinees ts)
  in

  let replace_scrutinees_in_tm open_up scrutinees ts tm =
    List.fold_left
      (fun tm (scrutinee, t) -> subst_fmeta (open_up scrutinee) t tm)
      tm (List.combine scrutinees ts)
  in

  let check_branchpattern env scrutinees typs pats = 
    ExtList.midfold
      (fun scrutinees_typs_pats (scrutinee,(typ,(metas,t))) (env, ncur, ts) ->
	 let nmetas = List.length metas in
	 let scrutinees, typs_pats = List.split scrutinees_typs_pats in
	 let env', _ = List.fold_left
	   (fun (env, i) (var,typ) ->
	      let open_up_in_env = open_up_meta (ncur + i) in
	      let typ' = open_up_in_env typ in
	      let _ = type_of_HOL env typ' in
	      let env' = add_to_env ~case:var typ' env in
		(env', i + 1))
	   (env, 0) metas
	 in
	 let open_up_in_env' = open_up_meta (ncur + nmetas) in
	 let t' = open_up_in_env' t in
	 let env'' = replace_scrutinees_in_env open_up_in_env' env' scrutinees ts in
	 let t_typ = type_of_HOL env'' t' in
	 let _ = check_pattern env' metas t in
	 let typ' = replace_scrutinees_in_tm open_up_in_env' scrutinees ts typ in
	   (if ctype_equal env' t_typ typ' then
	      ()
	    else
	      failwith "type of pattern is not the expected one");
	   (env', ncur + nmetas, List.append ts [open_up_in_env' t]))
      (List.combine scrutinees (List.combine typs pats))
      (env, 0, [])
  in

  let env', ncur, ts = check_branchpattern env scrutinees typs pats in
  let open_up_in_env' = open_up_meta ncur in
  let env'' = replace_scrutinees_in_env open_up_in_env' env' scrutinees ts in

  let e' = replace_scrutinees_in_tm open_up_in_env' scrutinees ts (open_up_in_env' e) in
  let e_typ = type_of_cterm env'' e' in
  let e_exp = ret_f ts in
    (if ctype_sub env' e_typ e_exp then
       ()
     else
       failwith "type of body of branch is not the expected one")

and check_ctxbranch ((defenv, cdefenv, metaenv, ctxenv, compenv) as env) scrutineevar ret_f (ctxvs, metas, t, e) =

  let open_up_ctx  n = CtxbindCterm.open_up ~howmany:n (List.length ctxenv) in
  let open_up_meta n = MetabindCterm.open_up ~howmany:n (List.length metaenv) in

  let env' = List.fold_left (fun env v -> add_to_env ~case:v (CSort(CCtx)) env) env ctxvs in
  let cn = List.length ctxvs in
  let metas' = ExtList.sndmap (open_up_ctx cn) metas in

  let env', n = List.fold_left (fun (env, n) (var, typ) ->
				  let typ' = open_up_meta n typ in
				  let _ = type_of_HOL env typ' in
				  let env' = add_to_env ~case:var typ' env in
				    (env', n+1)) (env',0) metas' in
  let t' = open_up_meta n (open_up_ctx cn t) in
  let _ = type_of_ctx env' t' in
  let _ = match t with CCtxTerm(LCtxAsList(l)) -> check_context_pattern env (List.length ctxenv) ctxvs metas l | _ -> failwith "only context lists allowed in context branches" in
  let e' = open_up_meta n (open_up_ctx cn e) in

  let env'' = subst_fctx_in_env env' (open_up_ctx cn scrutineevar) t' in
  let e'' = subst_fctx (open_up_ctx cn scrutineevar) t' e' in

  let e_typ = type_of_cterm env'' e'' in
  let e_exp = open_up_meta n (open_up_ctx cn (ret_f t)) in
    (if ctype_sub env' e_typ e_exp then
       ()
     else
       failwith "type of body of branch is not the expected one")

and env_from_subst defenv cdefenv (metasubst,ctxsubst,compsubst) =
  let metaenv = List.map (type_of_modal ([], defenv, [], [])) metasubst in
  let ctxenv  = List.map (ctx_wf ([], defenv, [], [])) ctxsubst in
  let compenv = List.map (type_of_cterm (defenv, cdefenv, [], [], [])) compsubst in
    (defenv, cdefenv, metaenv, ctxenv, compenv)

and type_of_HOL (defenv, cdefenv, metaenv, ctxenv, compenv) (e : cterm) =
  match e with
      CHolTerm(t) ->
	(let tp = Logic_typinf.type_of_modal ([], defenv, metaenv, ctxenv) t in CHolTerm(flatten_all_lmodal tp))
    | _ -> failwith "non-HOL term in HOL position"
and csort_of_HOL (defenv, cdefenv, metaenv, ctxenv, compenv) (e : cterm) =
  match e with
      CHolTerm(LTermInCtx(_,_) as t) ->
	(let _ = Logic_typinf.type_of_modal ([], defenv, metaenv, ctxenv) t in CHol)
    | CHolTerm(_) -> failwith "metavariables are not allowed as types of metavariables"
    | _ -> failwith "non-HOL term in HOL abstractor position"
and type_of_ctx (defenv, cdefenv, metaenv, ctxenv, compenv) (e : cterm) =
  match e with
      CCtxTerm(t) -> (let _ = Logic_typinf.ctx_wf ([], defenv, metaenv, ctxenv) t in CSort(CCtx))
    | _ -> failwith "non-context term in context position"
and csort_of_ctx (defenv, cdefenv, metaenv, ctxenv, compenv) (e : cterm) =
  match e with
      CSort(CCtx) -> CCtx
    | _ -> failwith "non-context term in context abstractor position"
and csort_of_cterm ((defenv, cdefenv, metaenv, ctxenv, compenv) as env) (e : cterm) =
  match ctype_fullwhdelta cdefenv (type_of_cterm env e) with
      CSort(CType) -> CType
    | CSort(CKind) -> CKind
    | _ -> failwith "non-allowed classifier in computational abstractor position"
and csort_of ~case:s env t = case_kind csort_of_HOL csort_of_ctx csort_of_cterm ~case:s env t
and type_of  ~case:s env t = case_kind type_of_HOL type_of_ctx type_of_cterm ~case:s env t

let type_of ~case:a b c = repeatmany 2 (fun () -> type_of ~case:a b c)

let rec cterm_def_add name tm tm_opt istrusted ?(expected_tp=None) defenv cdefenv =
  let env = (defenv, cdefenv, [], [], []) in
  let tp = type_of_cterm env tm in
  let add name what cdefenv =
    (if name = "_" then
	cdefenv
     else
	(if Dict.mem name cdefenv then
	    failwith ("definition " ^ name ^ " already exists")
	 else
	    Dict.add name what cdefenv))
  in
    match expected_tp with
	Some(t) ->
	  (if ctype_equal env tp t then
	     add name (t, tm, tm_opt, istrusted) cdefenv
	   else
	     failwith "the term's type doesn't match the definition type")
      | None ->
	  add name (tp, tm, tm_opt, istrusted) cdefenv

let cterm_def_add_trusted name tm tm_opt tp istrusted defenv cdefenv =
  let add name what cdefenv =
    (if name = "_" then cdefenv
     else
	(if Dict.mem name cdefenv then
	    failwith ("definition " ^ name ^ " already exists")
	 else
	   (*Format.fprintf Format.std_formatter "%s := %a : %a.@\n" name Comp_print.pr_cterm tm Comp_print.pr_cterm tp; *)
	    (Format.fprintf Format.std_formatter "compdefin %s : %a.@\n" name Comp_print.pr_cterm tp;
	     flush stdout;
	     Dict.add name what cdefenv)))
  in
    add name (tp, tm, tm_opt, istrusted) cdefenv

let comp_global_env : cterm_defenv ref = ref Dict.empty
let comp_global_reset () =
  let (d1,d2) = !Logic_main.logic_global_env in
    comp_global_env := Dict.empty;
    Logic_main.logic_global_env := (d1, Dict.empty)
let comp_global_reset_all () =
    comp_global_env := Dict.empty;
    Logic_main.logic_global_env := (Dict.empty, Dict.empty)
let comp_global_delete l =
  comp_global_env := (List.fold_left (fun d elm -> if Dict.mem elm d then Dict.remove elm d else d) !comp_global_env l)

let comp_global_save s =
  let f = open_out_bin (".compiled." ^ s ^ ".verimlc") in
    Marshal.to_channel f (!Logic_main.logic_global_env, !comp_global_env) [];
    close_out f

let comp_global_import s =
  let f = open_in_bin (".compiled." ^ s ^ ".verimlc") in
  let (logic, comp) = (Marshal.from_channel f : (Logic_ast.lterm_defenv * Comp_ast.cterm_defenv)) in
  let merge dexist d =
    Dict.fold
      (fun key v cur ->
	 if Dict.mem key cur then
	   cur
	 else
	   Dict.add key v cur)
      d
      dexist
  in
  let newlogic = (merge (fst !Logic_main.logic_global_env) (fst logic),
		  merge (snd !Logic_main.logic_global_env) (snd logic)) in
  let newcomp  = merge !comp_global_env comp in
    close_in f;
    Logic_main.logic_global_env := newlogic;
    comp_global_env := newcomp

(* let comp_gettype tm = *)
(*   let case = match tm with CHolTerm(_) -> (None, CHol) | CCtxTerm(_) -> (None, CCtx) | _ -> (None, CType) in *)
(*     type_of ~case:case (!Logic_main.logic_global_env, !comp_global_env, [], [], []) tm *)


(* after changes to comp_opteval for taking inferrence into account *)
(* let comp_gettype ?(expected=None) tm = *)
(*   let case = match tm with CHolTerm(_) -> (None, CHol) | CCtxTerm(_) -> (None, CCtx) | _ -> (None, CType) in *)
(*   let tp = type_of ~case:case (!Logic_main.logic_global_env, !comp_global_env, [], [], []) tm in *)
(*     match expected with *)
(* 	Some tp' -> *)
(* 	  (if ctype_equal (!Logic_main.logic_global_env, !comp_global_env, [], [], []) tp tp' then *)
(* 	     (tp, tm) *)
(* 	   else *)
(* 	     failwih "Type of computational term is not the expected one!") *)
(*       | None -> *)
(* 	  (tp, tm) *)

let comp_gettype ?(expected=None) tm =
  let case = match tm with CHolTerm(_) -> (None, CHol) | CCtxTerm(_) -> (None, CCtx) | _ -> (None, CType) in
  let tp = type_of ~case:case (!Logic_main.logic_global_env, !comp_global_env, [], [], []) tm in
    match expected with
	Some tp' ->
	  (if ctype_equal (!Logic_main.logic_global_env, !comp_global_env, [], [], []) tp tp' then
	     let tp = remove_infer_cterm tp in
	     let tm = remove_infer_cterm tm in
	       (tp, tm)
	   else
	     failwith "Type of computational term is not the expected one!")
      | None ->
	  let tp = remove_infer_cterm tp in
	  let tm = remove_infer_cterm tm in
	    (tp, tm)

let comp_print tm =
  match tm with
      CNVar(s) -> Comp_print.print_cterm (get_cdef_term s !comp_global_env)
    | _ -> Comp_print.print_cterm tm

let comp_print_type tm =
  Comp_print.print_cterm (fst (comp_gettype tm))
